"""Package metadata"""
TITLE = "aiocometd"
DESCRIPTION = "CometD client for asyncio"
KEYWORDS = "asyncio aiohttp comet cometd bayeux push streaming"
URL = "https://github.com/robertmrk/aiocometd"
VERSION = "0.2.2"
AUTHOR = "Róbert Márki"
AUTHOR_EMAIL = "gsmiko@gmail.com"
